comment #
*************************************************
*						*
*  PGAL controller   GAL16V8/20V8/22V10	*
*						*
*  (C)ChaN, 1995-1997				*
*						*
*************************************************

To re-assemble necessary OPTASM.
>optasm pgal; /ml

#

include	stddef.inc
include	str.inc
include	file.inc
include	ibmpc.inc

;-------------------------------
; Device-related definition

GAL16 	equ 1; device identification flag
GAL20 	equ 2
GAL22 	equ 3

ROW16 	equ 32; line number of addresses (16V8)
ROW20 	equ 40; number of input line (20V8)
ROW22 	equ 44; to the same (22V10)

AND16 	equ 64; AND the number of terms (16V8)
AND20 	equ 64; = line size (20V8)
AND22 	equ 132; (22V10)

F_ACW16 	equ ROW16 * AND16; ACW position of Fuse map (16V8)
F_ACW20 	equ ROW20 * AND20; (20V8)
F_ACW22 	equ ROW22 * AND22; (22V10)

FNUM16 	equ ROW16 * AND16 + 82 + 64; Fuse number (16V8)
FNUM20 	equ ROW20 * AND20 + 82 + 64; (20V8)
FNUM22 	equ ROW22 * AND22 + 20 + 64; (22V10)

ROW_ACW 	equ 60; ACW of the row address (16/20)
ROW_PRT 	equ 61; protected fuse row address (16/20)
ROW_ERA 	equ 63; line addresses of all erase fuse (16/20)


MESSAGE	macro	msc
		 mov	ah,msc
		 call	DISP_MES
		endm


.model t
.186
;-----------------------------------------------------------------------------
; Variable, the definition of the string area

.data

ER_NS 	equ 1; Unsupported Device
ER_WRT 	equ 2; write failure
ER_INI 	equ 3; PGAL is not connected

MS_JOB 	equ 4; in writing
MS_PASS 	equ 5; write completion
MS_OPN 	equ 6; start-up message
MS_USE 	equ 7; help message
MS_ERA 	equ 8; erase completion
MS_DEV 	equ 9; device type setting

MessErr		db	CR,LF,BEL,'Error: ',0
MessStr		db	ER_NS,	'%sNot supported device.%n',0
		db	ER_WRT,	'write failed.%n',BEL,0
		db	MS_ERA,	'%m                                                    %mErased.%n',0
		db	MS_JOB,	'%m                                                    %mWriting....',0
		db	ER_INI,	'%sPGAL not found.%n',0
		db	MS_PASS,'completed.%n',0
		db	MS_OPN,	'PGAL - GAL writer control program  R0.05 (C)ChaN,1995-1997%n%n',0
		db	MS_USE,	'Write : pgal <JEDEC file>%n'
		db		'Read  : pgal /R%n'
		db		'Erase : pgal /E%n',0
		db	MS_DEV, ' 1. GAL16V8%n'
		db		' 2. GAL20V8%n'
		db		' 3. GAL22V10%n'
		db		'%nEnter device type =',0
		db	-1,	'???',0

ErrJed		db	BEL,'JEDEC file error (%ui).%n',0
MessReq		db	'%mMount a "GAL%s" on the socket, and type any key.',0


;-------------------------------
; For JEDEC file manipulation

Type16		db	'16V8',0	; 16V8
Type20		db	'20V8',0	; 20V8
Type22		db	'22V10',0	; 22V10

Hd16		db	STX,CR,LF,'GAL16V8*QP20*QF2194*',CR,LF,'G0*F0*',CR,LF,0
Hd20		db	STX,CR,LF,'GAL20V8*QP24*QF2706*',CR,LF,'G0*F0*',CR,LF,0
Hd22		db	STX,CR,LF,'GAL22V10*QP24*QF5892*',CR,LF,'G0*F0*',CR,LF,0

HexChar		db	'0123456789ABCDEF'


;-------------------------------
; Control port-related

S_DAT		equ	BIT0	; bit allocation of the parallel output port
S_CLK		equ	BIT1	;
S_STB		equ	BIT2	;
T_TRG		equ	BIT3	;
G_CLK		equ	BIT4	;
G_DAT		equ	BIT5	;
G_STB		equ	BIT6	;

LPT_NUM		equ	1	; LPT # to be used in the IBM PC

CtrlPort	dw	?	; Address of the port
PortReg		db	?	;   of data

				; Of the shift register buffer and the bit allocation
				;     16V8   20V8   22V10
				; bit -----  -----  -----
Ctrl1		db	?	; b0: "1"    P/V    "0"     (U1-15)
				; b1: P/V    RA0    "0"     (U1-1)
				; b2: RA0    "0"    "0"     (U1-2)
				; b3: "0"    RA1    P/V     (U1-3)
				; b4: "0"    RA2    COM0    (U1-4)
				; b5: RA1    RA3    COM1    (U1-5)
				; b6: RA2    "0"    COM2    (U1-6)
				; b7: RA3    "0"    COM3    (U1-7)

Ctrl2		db	?	; b0: RA4    RA4    COM4    (U2-15)
				; b1: RA5    RA5    COM5    (U2-1)
				; b2: /VCC   "1"    "1"     (U2-2)
				; b3: "1"    /VCC   /VCC    (U2-3)
				; b4: /VPP   "1"    "1"     (U2-4)
				; b5: "1"    /VPP   /VPP    (U2-5)
				; b6: "1"    "0"    "1"     (U2-6)
				; b7: "0"    "1"    "0"     (U2-7)


;-------------------------------
; Device-related


Device		db	0		; Device type indicater
AcwLoc		dw	0		; ACW location in the fuse map
G_fuse		db	0		; Security fuse flag
Erase			db	0		; Erace only flag
FuseNum		dw	0		; Number of fuses

RowAddr		dw	?		; R/W row address (same as input line)
RowNum		dw	?		; Number of rows (same as number of input lines)
RowSize		dw	?		; Number of AND fuses par row
RowCount		dw	?		; Row data R/W buffer length
RowBuff		db	200 dup(?)	; Row data R/W buffer

FuseMap		db	6000 dup(?)	; Fuse data array


;-------------------------------
; File operations other

LineNum 	dw  	? 	;File read line number
CheckSum 	dw	?	;Checksum of the file; 

Hand 		dw	? 	;File handle
ReadPtr 	dw 	? 	;Buffer ptr
ReadBuff 	db 	256 dup (?)	 ; File read buffer
BlockBuff 	db 	8000 dup (?) ; File block buffer

Argc 		db	?
Argv 		dw	?




;-----------------------------------------------------------------------------
; Program code
;-----------------------------------------------------------------------------

.code

		DEF_PSP
start:
		cld
		MESSAGE  MS_OPN		; opening message
		GET_CMD_PAR  Argc,Argv	; command line acquisition
	_if ne, Argc, 1
		MESSAGE  MS_USE
		$DOS  4ch,-1
	_endif
		 call	INITIAL		; GAL writer initialization

		mov	bx,Argv[0]	; command analysis
		mov	ax,[bx]
		CAPS	ah
	_case e, ax,'R/'		; READ ?
		 call	READ_GAL
	_case e, ax,'E/'		; ERASE ?
		 call	ERASE_GAL
	_default			; else WRITE
		 call	WRITE_GAL
	_endc
		$DOS  4ch,0


dos_err:	ERR_OUT
		$DOS  4ch,1


;-----------------------------------------------------------------------------
; Fuse data read and display

READ_GAL	proc
		 call	REQ_INS
		 call	POW_ON
		 call	READ_FUSE	; Read fuse data
		 call	POW_OFF
		 call	DISP_JED	; Fuse Map ̕\
		ret
READ_GAL	endp


DISP_JED	proc
		mov	CheckSum,0
	_case e, Device,GAL16			; display of header
		mov	si,ofs Hd16
	_case e, Device,GAL20
		mov	si,ofs Hd20
	_default
		mov	si,ofs Hd22
	_endc
	_while
		lodsb
	 _break e, al,0
		 call	DISP
	_rept

		clr	si
	_repeat RowSize				; AND fuses
		push	cx
		mov	cx,RowNum
		 call	DP_MAP
		pop	cx
	_loop
	_if e, Device, GAL22			; GAL22
		mov	cx,20			; ACW
		 call	DP_MAP
		mov	cx,64			; Signature
		 call	DP_MAP
	_else					; GAL16/20
		mov	cx,8			; XOR
		 call	DP_MAP
		mov	cx,64			; Signature
		 call	DP_MAP
		mov	cx,8			; ACW1
		 call	DP_MAP
		mov	cx,64			; PTD
		 call	DP_MAP
		mov	cx,2			; SYN,AC0
		 call	DP_MAP
	_endif
		 call	DP_CSUM			; FUSE CHECK SUM
		 call	DP_DSUM			; FILE CHECK SUM
		ret
DISP_JED	endp


DP_MAP		proc			; Display of the fuse map
		mov	al,'L'
		 call	DISP
		mov	ax,si
		 call	DP_DEC
		mov	al,' '
		 call	DISP
	_repeat				; fuse data
		mov	al,FuseMap[si]
		inc	si
		add	al,'0'
		 call	DISP
	_loop
		 call	DP_DELIM
		ret
DP_MAP		endp

DP_CSUM		proc			; Checksum Export fuse map
		mov	al,'C'
		 call	DISP
		 call	GET_FSUM
		 call	DP_HEX		; check sum
		 call	DP_DELIM
		ret
DP_CSUM		endp


DP_DSUM		proc			; block the entire checksum writing
		mov	al,ETX
		 call	DISP
		mov	dx,CheckSum
		 call	DP_HEX
		mov	al,CR
		 call	DISP
		mov	al,LF
		 call	DISP
		ret
DP_DSUM		endp


GET_FSUM	proc			; Checksum calculation of the fuse
		clr	bx,dx
		mov	ax,0001h
	_repeat FuseNum
		tst	FuseMap[bx]
	 @ifn z,<add	dx,ax>
		inc	bx
		rol	al,1
	_loop
		ret
GET_FSUM	endp


DP_HEX		proc			; 16 hexadecimal output DX
		 call	dph
		 call	dph
		 call	dph
	dph:	rol	dx,4
		mov	al,dl
		and	al,0fh
		mov	bx,ofs HexChar
		xlat
		 call	DISP
		ret
DP_HEX		endp


DP_DEC		proc			; 10 decimal output AX
		push	cx
		mov	bp,10
	_repeat 4
		clr	dx
		div	bp
		push	dx
	_loop
	_repeat 4
		pop	ax
		add	al,'0'
		 call	DISP
	_loop
		pop	cx
		ret
DP_DEC		endp


DP_DELIM	proc			; delimiter Export
		mov	al,'*'
		 call	DISP
		mov	al,CR
		 call	DISP
		mov	al,LF
		 call	DISP
		ret
DP_DELIM	endp


DISP		proc			; character writing
		push	ax,dx
		clr	ah
		add	CheckSum,ax
		mov	dl,al
		$DOS  2
		pop	dx,ax
		ret
DISP		endp



;-----------------------------------------------------------------------------
; Fuse data erasure only

ERASE_GAL	proc
		 call	REQ_INS
		 call	POW_ON
		 call	ERASE_FUSE	; Erase fuse data
		 call	POW_OFF
		MESSAGE  MS_ERA
		ret
ERASE_GAL	endp



;-----------------------------------------------------------------------------
; Fuse data writing

WRITE_GAL	proc
		OPEN_FILE  0,Argv[0]	; open JED file
		jc	dos_err
		mov	Hand,ax
		 call	LOAD_JED	; load JED file
		CLOSE_FILE  Hand
		 call	REQ_INS
		MESSAGE  MS_JOB
		 call	POW_ON
		 call	ERASE_FUSE	; Erase fuse data
		 call	WRITE_FUSE	; Write fuse data
		 call	POW_OFF
		MESSAGE  MS_PASS
		ret
WRITE_GAL	endp


LOAD_JED	proc			; JEDEC file read
		mov	ReadPtr,0
		mov	LineNum,1
		 call	SKIP_HEAD	; header skip

	_while
		 call	READ_LINE	; 1 block read
	 _break e, al,EOF

		mov	bx,ofs BlockBuff
	 _while				; Skip top spaces
		mov	ax,[bx]
		inc	bx
	 _rept na, al,' '

	 _case e, al,'L'		; Fuse data ?
		 call	GET_VAL		; fuse location
		jc	lj_err
		mov	di,ax
	  _while			; set fuse map
		mov	al,[bx]
		inc	bx
	   _break e, al,'*'
	   _cont e, al,' '
		 call	chk_bit
		mov	FuseMap[di],al
		inc	di
	  _rept

	 _case e, ax,'AG'		; GAL type ?
		incw	bx
		CMP_STR  bx,<ofs Type16>,<size Type16-1>
	   @if e,<call	INIT_16>
		CMP_STR  bx,<ofs Type20>,<size Type20-1>
	   @if e,<call	INIT_20>
		CMP_STR  bx,<ofs Type22>,<size Type20-1>
	   @if e,<call	INIT_22>

	 _case e, al,'G'		; Security fuse ?
		mov	al,[bx]
		 call	chk_bit
		mov	G_fuse,al

	 _case e, ax,'FQ'		; Fuse number ?
		inc	bx
		 call	GET_VAL
		jc	lj_err
		mov	FuseNum,ax

	 _case e, al,'F'		; Fill map ?
		mov	al,[bx]
		 call	chk_bit
		FIL_STR  <ofs FuseMap>,<size FuseMap>

	 _case e, al,'C'		; Check sum ?
		 call	GET_VALH
		mov	di,ax
		 call	GET_FSUM
		cmp	dx,di
		jne	lj_err

	 _endc				; Other blocks (skip)

	_rept
		ret

		GET_VAL_F
		GET_VALH_F


chk_bit:	sub	al,'0'
		jb	lj_err
		cmp	al,1
		ja	lj_err
		ret

lj_err:		mov	bp,sp
		push	ofs ErrJed, LineNum
		 call	PRINTF
		$DOS  4ch,-1
LOAD_JED	endp


SKIP_HEAD	proc			; skip header
	_repeat
		 call	READ_CHR
	 _break	e, al,STX
	 _break	e, al,EOF
	 _if e, al,LF
		inc	LineNum
	 _endif
	_rept
		ret
SKIP_HEAD	endp

READ_LINE	proc			; 1 block read to BlockBuff
		clr	bx
	_repeat
		 call	READ_CHR
	 _break	e, al,EOF
	 _case e, al,CR
	 	mov	al,' '
	 _case e, al,LF
		inc	LineNum
	 	mov	al,' '
	 _endc
	 _cont b, al,' '
		mov	BlockBuff[bx],al
		inc	bx
	_rept ne, al,'*'
		mov	BlockBuff[bx],al
		mov	BlockBuff[bx][-1],' '
		ret
READ_LINE	endp


READ_CHR	proc			; One character read from the file
		push	bx,cx,dx
	_if e, ReadPtr,0
		READ_FILE Hand,<ofs ReadBuff>,<size ReadBuff>
		jc	dos_err
	 _if ne, ax,cx
		mov	bx,ax
		mov	ReadBuff[bx],EOF
	__endif
		mov	bx,ReadPtr
		mov	al,ReadBuff[bx]
		inc	bx
		cmp	bx,size ReadBuff
	@if ae,<clr	bx>
		mov	ReadPtr,bx
		pop	dx,cx,bx
		ret
READ_CHR	endp


;--------------------------------------;
; Message Output
;
; Call	ah = message code

DISP_MES	proc
		push	ax,si,bp
		mov	si,ofs MessStr
	_while
		lodsb
	 _break e, al,ah
	 _break e, al,-1
	 _repeat
		lodsb
	 _rept ne, al,0
	_rept
		mov	bp,sp
		push	si
		cmp	wptr[si],'s%'
	@if e, <push	ofs MessErr>
		 call	PRINTF
		pop	bp,si,ax
		ret
DISP_MES	endp

		PRINTF_F


;--------------------------------------;
; PGAL initialization

INITIAL		proc
		movseg	es,SEG_BIOS
		mov	ax,es:BiosLptLst[(LPT_NUM-1)*2]
		movseg	es,cs
	_skip e, ax,3bch
	 _skip e, ax,378h
	  _skip e, ax,278h
		mov	ax,40h
	___endif
		mov	CtrlPort,ax

		mov	Ctrl1,00000000b
		mov	Ctrl2,00111100b
		mov	PortReg,0000000b
		 call	SET_REG
		mov	bh,Ctrl2
		mov	bl,Ctrl1
		rol	bx,1
	_repeat 10
		mov	al,PortReg
		res	al,S_CLK
		outp	CtrlPort,al
		set	al,S_CLK+S_DAT
		out	dx,al
	 _if ne, dx,40h
		inc	dx
		in	al,dx
		shr	al,7
	 _else
		incw	dx
		in	al,dx
		shr	al,2
	 _endif
		and	al,1
		rol	bx,1
		xor	al,bl
		not	al
		test	al,BIT0
	 _break nz
		clr	al
	_loop
		mov	PortReg,0000000b
		outb	CtrlPort,0000000b
	_if nz
		MESSAGE ER_INI
		$DOS  4ch,1
	_endif
		FIL_STR  <ofs FuseMap>,<size FuseMap>,0
		ret
INITIAL		endp


;--------------------------------------;
; device power ON

POW_ON		proc

	_case e, Device, GAL16
		mov	Ctrl1,00000001b		; Vcc20 ON
		mov	Ctrl2,00111000b
		 call	SET_REG
		 call	DLY10
		set	PortReg,G_STB		; STR="H"
		mov	Ctrl2,01101000b		; Vpp20/pull up ON

	_case e, Device, GAL20
		mov	Ctrl1,00000000b		; Vcc24 ON
		mov	Ctrl2,00110100b
		 call	SET_REG
		 call	DLY10
		set	PortReg,G_STB		; STR="H"
		mov	Ctrl2,10010100b		; Vpp24/pull up ON

	_case e, Device, GAL22
		mov	Ctrl1,00000000b		; Vcc24 ON
		mov	Ctrl2,00110100b
		 call	SET_REG
		 call	DLY10
		set	PortReg,G_STB		; STR="H"
		mov	Ctrl2,01010100b		; Vpp24/pull up ON

	_endc
		 call	SET_REG
		 call	DLY10
		ret
POW_ON		endp


POW_RESET	proc			; device reset Vpp ? OFF ? ON
		mov	al,Ctrl2
		push	ax
		set	Ctrl2,BIT4+BIT5	; Vpp off
		 call	SET_REG
		 call	DLY10		; 10ms
		pop	ax
		mov	Ctrl2,al	; Vpp on
		 call	SET_REG
		 call	DLY10
		ret
POW_RESET	endp


INIT_16		proc		; Parameters set for GAL16V8
		mov	Device, GAL16
		mov	FuseNum, FNUM16
		mov	RowNum, ROW16
		mov	RowSize, AND16
		mov	AcwLoc, F_ACW16
		ret
INIT_16		endp

INIT_20		proc		; Parameters set for GAL20V8
		mov	Device, GAL20
		mov	FuseNum, FNUM20
		mov	RowNum, ROW20
		mov	RowSize, AND20
		mov	AcwLoc, F_ACW20
		ret
INIT_20		endp

INIT_22		proc		; Parameters set for GAL22V10
		mov	Device, GAL22
		mov	FuseNum, FNUM22
		mov	RowNum, ROW22
		mov	RowSize, AND22
		mov	AcwLoc, F_ACW22
		ret
INIT_22		endp



;--------------------------------------;
; device power OFF

POW_OFF		proc
		set	Ctrl2,BIT4+BIT5	; Vpp off
		 call	SET_REG
		 call	DLY
		 call	DLY
		 call	DLY
		 call	DLY
		mov	Ctrl1,00000000b	; Vcc off
		mov	Ctrl2,00111100b
		mov	PortReg,0000000b
		 call	SET_REG
		ret
POW_OFF		endp


;--------------------------------------;
; device set request

REQ_INS		proc
	_if e, Device,0
		mov	ah,MS_DEV
		 call	DISP_MES
	 _while
		$DOS  8
	  _case e, al,'1'
		 call	INIT_16
	   _break
	  _case e, al,'2'
		 call	INIT_20
	   _break
	  _case e, al,'3'
		 call	INIT_22
	   _break
	  _endc
	 _rept
	_endif
		mov	bp,sp
		push	ofs MessReq
	_case e, Device, GAL16
		push	ofs Type16
	_case e, Device, GAL20
		push	ofs Type20
	_default
		push	ofs Type22
	_endc
		 call	PRINTF
		$DOS  8
		ret
REQ_INS		endp


;--------------------------------------;
; fuse data read
;
; Ret	FuseMap = read data

READ_FUSE	proc
		mov	RowAddr,0		; Read product lines
		mov	ax,RowSize
		mov	RowCount,ax
	_repeat RowNum
		push	cx
		 call	READ_ROW
		mov	bx,RowAddr
		mov	si,ofs RowBuff
	 _repeat RowSize
		lodsb
		mov	FuseMap[bx],al
		add	bx,RowNum
	 _loop
		inc	RowAddr
		pop	cx
	_loop

		mov	RowCount,64
	_if e, Device, GAL22
		 call	READ_ROW		; Read signature
		MOV_MEM <ofs RowBuff[0]>,<ofs FuseMap[5828]>,64
		 call	READ_ACW22
		clr	bx
	 _repeat 10
		mov	ax,wptr RowBuff[bx]
		xchg	ah,al
		mov	wptr FuseMap[F_ACW22+bx],ax
		incw	bx
	 _loop

	_else		;GAL16/20
		addi	di,<ofs FuseMap>,AcwLoc
		 call	READ_ROW		; Read signature
		add	di,8
		MOV_MEM <ofs RowBuff[0]>,,64
		sub	di,72
		mov	RowAddr,ROW_ACW		; read ACW
		mov	RowCount,82
		 call	READ_ROW
		MOV_MEM <ofs RowBuff[ 0]>,, 4	; XOR
		MOV_MEM <ofs RowBuff[78]>,, 4	; XOR
		add	di,64			; signature area
		MOV_MEM <ofs RowBuff[ 5]>,, 4	; AC1
		MOV_MEM <ofs RowBuff[73]>,, 4	; AC1
		MOV_MEM <ofs RowBuff[ 9]>,,64	; PTD
		MOV_MEM <ofs RowBuff[77]>,, 1	; SYN
		MOV_MEM <ofs RowBuff[ 4]>,, 1	; AC0
	_endif
		ret
READ_FUSE	endp


;--------------------------------------;
; erasure of fuse data
;

ERASE_FUSE	proc
	_if e, Device, GAL22			; Erase GAL22
		set	Ctrl1,11010000b	; COM=3Dh
		set	Ctrl2,00000011b
		 call	SET_REG
		 call	STB_WRITE	; STB
		res	Ctrl1,11110000b	; COM=00h
		res	Ctrl2,00000011b
		 call	SET_REG
	_else					; Erase GAL16/20
		mov	RowAddr,ROW_ERA
		mov	RowCount,1
		mov	RowBuff[0],1
		 call	WRITE_ROW
	_endif
		 call	POW_RESET
		ret
ERASE_FUSE	endp


;--------------------------------------;
; writing of fuse data
;
; Call	FuseMap = write data

WRITE_FUSE	proc
		mov	RowAddr,0		; Write product fuses
		mov	ax,RowSize
		mov	RowCount,ax
	_repeat RowNum
		push	cx
		mov	bx,RowAddr
		mov	di,ofs RowBuff
	 _repeat RowSize
		mov	al,FuseMap[bx]
		add	bx,RowNum
		stosb
	 _loop
		 call	WRITE_ROW
		jne	write_fail
		inc	RowAddr
		pop	cx
	_loop

		mov	RowCount,64
	_if e, Device, GAL22			; 22V10 signature and ACW
		MOV_MEM <ofs FuseMap[5828]>,<ofs RowBuff[0]>,64
		 call	WRITE_ROW
		jne	write_fail
		clr	bx
	 _repeat 10
		mov	ax,wptr FuseMap[F_ACW22+bx]
		xchg	ah,al
		mov	wptr RowBuff[bx],ax
		incw	bx
	 _loop
		 call	WRITE_ACW22
		jne	write_fail

	_else					; GAL16/20 signature and ACW
		lea	si,FuseMap[bx+9]
		sub	si,RowNum
		MOV_MEM <>,<ofs RowBuff[0]>,64	; Write signature
		 call	WRITE_ROW
		jne	write_fail
		sub	si,72			; Write ACW
		MOV_MEM <>,<ofs RowBuff[ 0]>, 4	; XOR
		MOV_MEM <>,<ofs RowBuff[78]>, 4	; XOR
		add	si,64			; signature area
		MOV_MEM <>,<ofs RowBuff[ 5]>, 4	; AC1
		MOV_MEM <>,<ofs RowBuff[73]>, 4	; AC1
		MOV_MEM <>,<ofs RowBuff[ 9]>,64	; PTD
		MOV_MEM <>,<ofs RowBuff[77]>, 1	; SYN
		MOV_MEM <>,<ofs RowBuff[ 4]>, 1	; AC0
		mov	RowAddr,ROW_ACW
		mov	RowCount,82
		 call	WRITE_ROW
		jne	write_fail
	_endif

	_if nz,<tst G_fuse>			; Security fuse ?
		mov	RowAddr,ROW_PRT
		mov	RowCount,1
		mov	RowBuff[0],1
		 call	WRITE_ROW
	_endif
		ret

write_fail:	 call	POW_OFF
		MESSAGE  ER_WRT
		$DOS  4ch,-1
WRITE_FUSE	endp


;--------------------------------------;
; Line reading
;
; Call	RowAddr = Row address
;	RowCount = Size of R/W length
;
; Ret	RowBuff = read data

READ_ROW	proc
		push	di
		 call	SET_ADDR	; set row address
		 call	SFT_ADDR
		 call	STB_READ	; read latch
		mov	di,ofs RowBuff	; read from shift reg.
	_repeat RowCount
		 call	READ_BIT
		stosb
	_loop
		pop	di
		ret
READ_ROW	endp

		; 22V10 ACW reading
READ_ACW22	proc
		set	Ctrl2,BIT0	; CM4 = H
		 call	SET_REG
		 call	STB_READ	; read latch
		mov	di,ofs RowBuff	; read from shift reg.
	_repeat 20
		 call	READ_BIT
		stosb
	_loop
		res	Ctrl2,BIT0	; CM4 = L
		 call	SET_REG
		ret
READ_ACW22	endp


;--------------------------------------;
; row write
;
; Call	RowBuff = write data
;	RowCount = Row length
; Ret	ZF = 1:pass, 0:fail

WRITE_ROW	proc
		push	si
		 call	SET_ADDR	; set row address
		mov	si,ofs RowBuff 	; send to shift reg.
	_repeat RowCount
		lodsb
		 call	SFT_BIT
	_loop
		 call	SFT_ADDR
		 call	STB_WRITE	; program
		 call	STB_READ	; read latch
		clr	si	 	; compere to row buffer
	_repeat RowCount
		 call	READ_BIT
	 _break ne, al,RowBuff[si]
		inc	si
		clr	al
	_loop
		pop	si
		ret
WRITE_ROW	endp


		; 22V10 ACW writing
WRITE_ACW22	proc
		set	Ctrl2,BIT0	; CM4 = H
		 call	SET_REG
		mov	si,ofs RowBuff 	; send to shift reg.
	_repeat 20
		lodsb
		 call	SFT_BIT
	_loop
		 call	STB_WRITE	; program
		 call	STB_READ	; read latch
		clr	si	 	; compere to row buffer
	_repeat 20
		 call	READ_BIT
	 _break ne, al,RowBuff[si]
		inc	si
		clr	al
	_loop
		pushf
		res	Ctrl2,BIT0	; CM4 = L
		 call	SET_REG
		popf
		ret
WRITE_ACW22	endp


;--------------------------------------;
; Row address setting
;
; Call	RowAddr = row address (0..63)

SET_ADDR	proc
		mov	ah,bptr RowAddr
		mov	al,Ctrl1
	_case e, Device,GAL16		; GAL16
		shl	al,6
		shr	ax,1	; RA0
		shr	al,2	; L,L
		shr	ax,3	; RA1,2,3
	_case e, Device,GAL20		; GAL20
		shl	al,7
		shr	ax,1	; RA0
		shr	al,1	; L
		shr	ax,3	; RA1,2,3
		shr	al,2	; L,L
	_default			; GAL22
		ret
	_endc
		mov	Ctrl1,al
		mov	al,Ctrl2
		and	al,11111100b
		and	ah,00000011b
		or	al,ah
		mov	Ctrl2,al
		 call	SET_REG
		ret
SET_ADDR	endp


;--------------------------------------;
; Read / Write Strobe

STB_READ	proc
	_case e, Device, GAL16			; set P/V = L
		res	Ctrl1,BIT1
	_case e, Device, GAL20
		res	Ctrl1,BIT0
	_default	;GAL22
		res	Ctrl1,BIT3
	_endc
		 call	SET_REG
		res	PortReg,G_STB		; STB = L
		outb	CtrlPort,PortReg
		 call	DLY
		set	PortReg,G_STB		; STB = H
		outb	CtrlPort,PortReg
		 call	DLY
		ret
STB_READ	endp

STB_WRITE	proc
	_case e, Device, GAL16			; set P/V = H
		set	Ctrl1,BIT1
	_case e, Device, GAL20
		set	Ctrl1,BIT0
	_default	;GAL22
		set	Ctrl1,BIT3
	_endc
		 call	SET_REG
		res	PortReg,G_STB		; STB = L
		outb	CtrlPort,PortReg
		 call	DLY10			; Delay for 40ms
		 call	DLY10
		 call	DLY10
		 call	DLY10
		set	PortReg,G_STB		; STB = H
		outb	CtrlPort,PortReg
		 call	DLY
		ret
STB_WRITE	endp


;--------------------------------------;
; 1 bit read
;
; Ret	AL = 0/1

READ_BIT	proc
		push	dx
	_if e, Device,  GAL20
		set	PortReg,S_DAT
		res	PortReg,S_CLK
	_else		;16/22
		res	PortReg,S_DAT+S_CLK
	_endif
		outb	CtrlPort,PortReg
		call	DLY
	_if ne, dx,40h		; IBMPC
		inc	dx
		mov	ah,BIT7
	_else			; PC98
		incw	dx
		mov	ah,BIT2
	_endif
		in	al,dx
		test	al,ah
		mov	ah,0
	@if z, <mov	ah,1>
		 call	CLK
		mov	al,ah
		pop	dx
		ret
READ_BIT	endp


;--------------------------------------;
; GAL data set
;
; Call	AL.0 = 0/1

SFT_BIT		proc
		push	ax,dx
	_if z, <test	al,BIT0>
		res	PortReg,G_DAT
	_else
		set	PortReg,G_DAT
	_endif
		outb	CtrlPort,PortReg
		 call	DLY
		 call	CLK
		pop	dx,ax
		ret
SFT_BIT		endp


SFT_ADDR	proc
	_if e, Device,GAL22
		mov	ax,RowAddr
	 _repeat 6
		 call	SFT_BIT
		shr	al,1
	 _loop
		res	PortReg,G_DAT
		outb	CtrlPort,PortReg
		 call	DLY
	_endif
		ret
SFT_ADDR	endp


;--------------------------------------;
; GAL shift clock

CLK		proc			; GAL shift clock
		mov	al,PortReg
		set	al,G_CLK
		outp	CtrlPort,al
		 call	DLY
		res	al,G_CLK
		outp	CtrlPort,al
		 call	DLY
		ret
CLK		endp


;--------------------------------------;
; Setting of PGAL of the shift register
;
; Ctrl1, to transfer the value of the Ctrl2

SET_REG		proc
		push	ax,bx,cx,dx
		mov	dx,CtrlPort
		mov	al,PortReg
		mov	bh,Ctrl2	; send U1/U2 data
		mov	bl,Ctrl1
	_repeat 16
		res	al,S_DAT+S_CLK
		shl	bx,1
	 @if c,<set	al,S_DAT>
		out	dx,al
		set	al,S_CLK
		out	dx,al
	_loop
		set	al,S_STB	; STB
		out	dx,al		;
		res	al,S_STB+S_CLK	;
		out	dx,al		;
		mov	PortReg,al
		pop	dx,cx,bx,ax
		 call	DLY
		ret
SET_REG		endp


;--------------------------------------;
; 10 ms waiting

DLY10		proc
		push	ax,dx
		set	PortReg,T_TRG+S_CLK	; trigger timer
		outb	CtrlPort,PortReg
		res	PortReg,T_TRG+S_DAT
		outb	CtrlPort,PortReg

	_if ne, dx,40h		; IBMPC
		inc	dx
		mov	ah,BIT7
	_else			; PC98
		mov	dx,42h
		mov	ah,BIT2
	_endif
	_while
		in	al,dx
	_rept z,<test	al,ah>
		pop	dx,ax
		ret
DLY10		endp


;--------------------------------------;
; Number s wait (I / O cycle use)

DLY		proc
		push	ax,dx
		inp	al,CtrlPort
		in	al,dx
		in	al,dx
		in	al,dx
		in	al,dx
		pop	dx,ax
		ret
DLY		endp


		$end	start
