comment #
*************************************************
*						*
* PGAL for PIC16F84 writing program *
*						*
*  (C)ChaN, 1997				*
*						*
*************************************************
#

include	stddef.inc
include	str.inc
include	file.inc
include	ibmpc.inc


MESSAGE		macro	msc
		mov	ah,msc
		 call	DISP_MES
		endm

SEND_CMD	macro	cmd
		mov	al,cmd
		 call	XMIT_CMD
		endm

LOAD_CFG	equ	000000b	; Command
LOAD_PRG	equ	000010b
READ_PRG	equ	000100b
INCR_ADR	equ	000110b
BEGIN_WR	equ	001000b
LOAD_DAT	equ	000011b
READ_DAT	equ	000101b
ERA_PROG	equ	001001b
ERA_DATA	equ	001011b


P_SIZE		equ	1024	; Size of program memory
D_SIZE		equ	64	; Size of data memory


.model tiny
.186
;-----------------------------------------------------------------------------
; Variable, the definition of the string area

.data

ER_FAIL 	equ 2; verify error
ER_INI 	equ 1; PGAL is not connected
ER_HEX 	equ 4 ;Invalid HEX file; 
ER_NCFG 	equ 5 ;There is no CFG data; 

MS_JOB1 	equ 20; CODE in writing
MS_JOB2 	equ 21; DATA in writing
MS_JOB3 	equ 22; during the CFG writing
MS_PASS 	equ 10; write completion
MS_ERA 	equ 11; erase completion

MS_OPN 	equ 12; boot message
MS_USE 	equ 13; help message

MessStr		db	ER_FAIL,'failed.%n',BEL,0
		db	MS_PASS,'pass.%n',0
		db	ER_HEX,	'Hex file error.%n',BEL,0
		db	ER_NCFG,'WARNING: Configuration word not defined.%n',0
		db	MS_ERA,	'Erased.%n',0
		db	MS_JOB1,'Writing program code....',0
		db	MS_JOB2,'Writing data memory....',0
		db	MS_JOB3,'Writing ID and cofiguration....',0
		db	ER_INI,	'PGAL not found.%n',0
		db	MS_OPN,	'PICW84 - PIC16x84 read/write controller for PGAL.%n         R0.01 (C)ChaN,1997%n%n',0
		db	MS_USE,	'Write : picw84 <HEX file>%n'
		db		'Read  : picw84 /R%n'
		db		'Erase : picw84 /E%n',0
		db	-1,	'???',0



;-------------------------------
; Control port-related

S_DAT 	equ 	BIT0	; bit allocation of the parallel output port
S_CLK		equ	BIT1	;
S_STB		equ	BIT2	;
T_TRG		equ	BIT3	;
P_CLK		equ	BIT4	;
P_DAT		equ	BIT5	;

LPT_NUM		equ	1	; LPT # to be used in the IBM PC

CtrlPort	dw	?	; Address of the port
PortReg	db	?	;                data

				; shift register buffer and the bit allocation
Ctrl1		db	?	; b0: "0"     (U1-15)
				; b1: "0"     (U1-1)
				; b2: "0"     (U1-2)
				; b3: "0"     (U1-3)
				; b4: "0"     (U1-4)
				; b5: "0"     (U1-5)
				; b6: "0"     (U1-6)
				; b7: "0"     (U1-7)

Ctrl2		db	?	; b0: "0"     (U2-15)
				; b1: "0"     (U2-1)
				; b2: "1"     (U2-2)
				; b3: /VCC    (U2-3)
				; b4: "1"     (U2-4)
				; b5: /VPP    (U2-5)
				; b6: "1"     (U2-6)
				; b7: "0"     (U2-7)


;-------------------------------
; Data buffer

ProgFlag	db	OFF		; Program
CfgFlag		db	OFF		; control data
DataFlag	db	OFF		; EEPROM data

ProgBuff	dw	P_SIZE dup(?)	; Program 
CfgBuff		dw	8      dup(?)	; Device control
DataBuff	dw	D_SIZE dup(?)	; EEPROM data


;-------------------------------
; other file operations

Hand		dw	?		; file handle
ReadPtr		dw	?		; buffer ptr
ReadBuff	db	256 dup(?)	; file read buffer
BlockBuff	db	260 dup(?)	; HEX block buffer

Argc		db	?
Argv		dw	?



;-----------------------------------------------------------------------------
; Program code
;-----------------------------------------------------------------------------

.code

		DEF_PSP
start:
		cld
		MESSAGE  MS_OPN		; opening message
		GET_CMD_PAR  Argc,Argv	; get command line
	_if ne, Argc, 1
		MESSAGE  MS_USE
		$DOS  4ch,-1
	_endif
		 call	INITIAL		; PGAL initialization

		mov	bx,Argv[0]	; analyse command
		mov	ax,[bx]
		CAPS	ah
	_case e, ax,'R/'		; READ ?
		 call	READ_PIC
	_case e, ax,'E/'		; ERASE ?
		 call	ERASE_PIC
	_default			; else WRITE
		 call	WRITE_PIC
	_endc
		$DOS  4ch,0


dos_err:	ERR_OUT
		$DOS  4ch,1


;--------------------------------------;
; message output
;
; Call	ah = message code

DISP_MES	proc
		push	ax,si,bp
		mov	si,ofs MessStr
	_while
		lodsb
	 _break e, al,ah
	 _break e, al,-1
	 _repeat
		lodsb
	 _rept ne, al,0
	_rept
		mov	bp,sp
		push	si
		 call	PRINTF
		pop	bp,si,ax
		ret
DISP_MES	endp

		PRINTF_F


;-----------------------------------------------------------------------------
; Program data read and display

READ_PIC	proc
		 call	VCC_ON		; Vcc ON
		 call	READ_PROG	; Read program and config.
		 call	READ_DATA	; Read data.
		 call	VCC_OFF		; Vcc OFF
		 call	DISP_HEX	; HEX display
		ret
READ_PIC	endp


DISP_HEX	proc
		mov	si,ofs ProgBuff		; Program blocks
		mov	bx,0
	_repeat P_SIZE*2/32
		push	cx
		mov	cx,32
		 call	DP_BLK
		pop	cx
	_loop
		mov	si,ofs CfgBuff		; ID block
		mov	bx,2000h*2
		mov	cx,4*2
		 call	DP_BLK
		mov	si,ofs CfgBuff[7*2]	; Config block
		mov	bx,(2000h+7)*2
		mov	cx,1*2
		 call	DP_BLK
		mov	si,ofs DataBuff		; Data blocks
		mov	bx,2100h*2
	_repeat D_SIZE*2/32
		push	cx
		mov	cx,32
		 call	DP_BLK
		pop	cx
	_loop
		ret
DISP_HEX	endp


DP_BLK		proc
		mov	dl,':'
		$DOS  2
		clr	dl
		mov	al,cl	; Block Len.
		 call	DP_HEX
		mov	al,bh	; Addr.H
		 call	DP_HEX
		mov	al,bl	; Addr.L
		 call	DP_HEX
		mov	al,0	; Block Type
		 call	DP_HEX
	_repeat			; Data
		lodsb
		 call	DP_HEX
		inc	bx
	_loop
		mov	al,dl	; Sum
		neg	al
		 call	DP_HEX
		mov	dl,CR	; CR
		$DOS  2
		mov	dl,LF	; LF
		$DOS  2
		ret
DP_BLK		endp


DP_HEX		proc			; 16 hexadecimal output AL
		add	dl,al
		call	dph
	dph:	rol	al,4
		push	ax,dx
		and	al,0fh
		cmp	al,10
	@if ae,<add	al,7>
		add	al,'0'
		mov	dl,al
		$DOS  2
		pop	dx,ax
		ret
DP_HEX		endp



;-----------------------------------------------------------------------------
; Erase only

ERASE_PIC	proc
		 call	VCC_ON		; Vcc ON
		 call	ERASE_ALL	; Erase all data.
		 call	VCC_OFF		; Vcc OFF
		MESSAGE  MS_ERA
		ret
ERASE_PIC	endp



;-----------------------------------------------------------------------------
; writing

WRITE_PIC	proc
		OPEN_FILE  0,Argv[0]	; open HEX file
		jc	dos_err
		mov	Hand,ax
		mov	ReadPtr,0
		 call	LOAD_HEX	; load HEX file
		CLOSE_FILE  Hand

		 call	VCC_ON		; Power ON

	_if e, ProgFlag,ON
		MESSAGE  MS_JOB1
		 call	WRITE_PROG	; Write program code
		MESSAGE  MS_PASS
	_endif

	_if e, DataFlag,ON		; Write EEPROM
		MESSAGE  MS_JOB2
		 call	WRITE_DATA
		MESSAGE  MS_PASS
	_endif

	_if e, ProgFlag,ON
	 _if ne, CfgFlag,ON
		MESSAGE  ER_NCFG
	 _endif
		MESSAGE  MS_JOB3
		 call	WRITE_CFG	; Write CFG word
		MESSAGE  MS_PASS
	_endif

		 call	VCC_OFF		; Power OFF
		ret
WRITE_PIC	endp


LOAD_HEX	proc			; HEX file read
	_while
		 call	READ_BLOCK	; 1 block read
	 _break c
		mov	si,ofs BlockBuff
	 _if e, bptr[si][3],0	; data block ?
	  _if ne, bptr[si][0],0
		clr	cx
		mov	cl,[si][0]
		mov	bh,[si][1]
		mov	bl,[si][2]
		add	si,4
	   _repeat
		lodsb
	    _case ae, bx, 2100h*2
	     _skip ae, bx, (2100h+D_SIZE)*2
		mov	bptr DataBuff[bx-2100h*2],al
		set	DataFlag
	     _endif
	    _case ae, bx, 2000h*2
	     _skip ae, bx, 2008h*2
		mov	bptr CfgBuff[bx-2000h*2],al
	      _skip ne, bx, 2007h*2
		set	CfgFlag
	     __endif
	    _default
	     _skip ae, bx, P_SIZE*2
		mov	bptr ProgBuff[bx],al
		set	ProgFlag
	     _endif
	    _endc
		inc	bx
	   _loop
	 __endif
	_rept
		ret
LOAD_HEX	endp


READ_BLOCK	proc			; 1 block read to BlockBuff
	_while
		 call	READ_CHR
	 _if e, al,EOF
		stc
		ret
	 _endif
	_rept ne, al,':'
		mov	di,ofs BlockBuff
		 call	READ_BYTE
		stosb
		mov	dl,al
		mov	cl,al
		clr	ch
		add	cx,4
	_repeat
		 call	READ_BYTE
		stosb
		add	dl,al
	_loop
		cmp	dl,0
		jne	lh_err
		ret
READ_BLOCK	endp


READ_BYTE	proc			; 1 byte read
		 call	rb_n
		shl	al,4
		mov	bl,al
		 call	rb_n
		or	al,bl
		ret
rb_n:
		 call	READ_CHR
		CAPS	al
		sub	al,'0'
	_if ae, al,10
		sub	al,7
		cmp	al,16
		jae	lh_err
		cmp	al,10
		jnae	lh_err
	_endif
		ret
READ_BYTE	endp


READ_CHR	proc			; 1 char read
		push	bx,cx,dx
	_if e, ReadPtr,0
		READ_FILE Hand,<ofs ReadBuff>,<size ReadBuff>
		jc	dos_err
	 _if ne, ax,cx
		mov	bx,ax
		mov	ReadBuff[bx],EOF
	__endif
		mov	bx,ReadPtr
		mov	al,ReadBuff[bx]
		inc	bx
		cmp	bx,size ReadBuff
	@if ae,<clr	bx>
		mov	ReadPtr,bx
		pop	dx,cx,bx
		ret
READ_CHR	endp

lh_err:		MESSAGE  ER_HEX
		$DOS  4ch,-1



;--------------------------------------;
; program / data read
;

READ_PROG	proc
		 call	VPP_ON

		mov	di,ofs ProgBuff
	_repeat P_SIZE
		SEND_CMD  READ_PRG
		 call	RCVR_DATA
		stosw
		SEND_CMD  INCR_ADR
	_loop

		SEND_CMD  LOAD_CFG
		mov	ax,-1
		 call	XMIT_DATA
		mov	di,ofs CfgBuff
	_repeat 4
		SEND_CMD  READ_PRG
		 call	RCVR_DATA
		stosw
		SEND_CMD  INCR_ADR
	_loop
		SEND_CMD  INCR_ADR
		SEND_CMD  INCR_ADR
		SEND_CMD  INCR_ADR
		SEND_CMD  READ_PRG
		 call	RCVR_DATA
		mov	[di][6],ax

		 call	VPP_OFF
		ret
READ_PROG	endp


READ_DATA	proc
		 call	VPP_ON
		mov	di,ofs DataBuff
	_repeat D_SIZE
		SEND_CMD  READ_DAT
		 call	RCVR_DATA
		clr	ah
		stosw
		SEND_CMD  INCR_ADR
	_loop
		 call	VPP_OFF
		ret
READ_DATA	endp


;--------------------------------------;
; chip erase
;

ERASE_ALL	proc
		 call	VPP_ON
		SEND_CMD  LOAD_CFG
		mov	ax,-1
		 call	XMIT_DATA
	_repeat 7
		SEND_CMD  INCR_ADR
	_loop
		SEND_CMD  000001b
		SEND_CMD  000111b
		SEND_CMD  BEGIN_WR
		 call	DLY10
		SEND_CMD  000001b
		SEND_CMD  000111b
		 call	VPP_OFF
		ret
ERASE_ALL	endp


;--------------------------------------;
; Program / data writing
;

WRITE_PROG	proc
		 call	VPP_ON
		mov	si,ofs ProgBuff	; Program
	_repeat P_SIZE
		 call	WP_PROG
	_loop
		 call	VPP_OFF
		ret
WRITE_PROG	endp

WRITE_CFG	proc
		 call	VPP_ON

		SEND_CMD  LOAD_CFG	; ID
		mov	ax,-1
		 call	XMIT_DATA
		mov	si,ofs CfgBuff
	_repeat 4
		 call	WP_PROG
	_loop
		SEND_CMD  INCR_ADR	; Config
		SEND_CMD  INCR_ADR
		SEND_CMD  INCR_ADR
		add	si,6
		 call	WP_PROG

		 call	VPP_OFF
		ret
WRITE_CFG	endp


WP_PROG		proc
		SEND_CMD  READ_PRG
		 call	RCVR_DATA
		xor	ax,[si]
		test	ax,3fffh
	 _skip z
		SEND_CMD  LOAD_PRG
		mov	ax,[si]
		 call	XMIT_DATA
		SEND_CMD  BEGIN_WR
		 call	DLY10
		SEND_CMD  READ_PRG
		 call	RCVR_DATA
		xor	ax,[si]
		and	ax,3fffh
		jne	write_fail
	 _endif
		SEND_CMD  INCR_ADR
		incw	si
		ret
WP_PROG		endp


WRITE_DATA	proc
		 call	VPP_ON

		mov	si,ofs DataBuff
	_repeat D_SIZE
		SEND_CMD  READ_DAT
		 call	RCVR_DATA
	 _skip e, al,[si]
		SEND_CMD  LOAD_DAT
		mov	ax,[si]
		 call	XMIT_DATA
		SEND_CMD  BEGIN_WR
		 call	DLY10
		SEND_CMD  READ_DAT
		 call	RCVR_DATA
		cmp	al,[si]
		jne	write_fail
	 _endif
		SEND_CMD  INCR_ADR
		incw	si
	_loop

		 call	VPP_OFF
		ret
WRITE_DATA	endp


write_fail:	 call	VCC_OFF
		MESSAGE  ER_FAIL
		$DOS  4ch,-1



;--------------------------------------;
; PGAL initialization

INITIAL		proc
		movseg	es,SEG_BIOS
		mov	ax,es:BiosLptLst[(LPT_NUM-1)*2]
		movseg	es,cs
	_skip e, ax,3bch
	 _skip e, ax,378h
	  _skip e, ax,278h
		mov	ax,40h
	___endif
		mov	CtrlPort,ax

		mov	Ctrl1,00000000b
		mov	Ctrl2,00111100b
		mov	PortReg,0000000b
		 call	SET_REG
		mov	bh,Ctrl2
		mov	bl,Ctrl1
		rol	bx,1
	_repeat 10
		mov	al,PortReg
		res	al,S_CLK
		outp	CtrlPort,al
		set	al,S_CLK+S_DAT
		out	dx,al
	 _if ne, dx,40h
		inc	dx
		in	al,dx
		shr	al,7
	 _else
		incw	dx
		in	al,dx
		shr	al,2
	 _endif
		and	al,1
		rol	bx,1
		xor	al,bl
		not	al
		test	al,BIT0
	 _break nz
		clr	al
	_loop
		mov	PortReg,0000000b
		outb	CtrlPort,0000000b
	_if nz
		MESSAGE ER_INI
		$DOS  4ch,1
	_endif
		FIL_STR  <ofs ProgBuff>,<size ProgBuff>,-1
		FIL_STR  <ofs CfgBuff>,<size CfgBuff>,-1
		FIL_STR  <ofs DataBuff>,<size DataBuff>,-1
		ret
INITIAL		endp



;--------------------------------------;
; Command transmission

XMIT_CMD	proc
		push	cx,dx
		mov	ah,al
		mov	al,PortReg
	_repeat 6
		set	al,P_CLK+P_DAT
		shr	ah,1
	 @if c,<res	al,P_DAT>
		outp	CtrlPort,al
		 call	DLY
		res	al,P_CLK
		outp	CtrlPort,al
	_loop
		 call	DLY
		pop	dx,cx
		ret
XMIT_CMD	endp


;--------------------------------------;
; data transmission

XMIT_DATA	proc
		push	ax,bx,cx,dx
		mov	bx,ax
		shl	bx,1
		and	bx,0111111111111110b
		mov	al,PortReg
	_repeat 16
		set	al,P_CLK+P_DAT
		shr	bx,1
	 @if c,<res	al,P_DAT>
		outp	CtrlPort,al
		 call	DLY
		res	al,P_CLK
		outp	CtrlPort,al
	_loop
		pop	dx,cx,bx,ax
		ret
XMIT_DATA	endp


;--------------------------------------;
; data reception

RCVR_DATA	proc
		push	bx,cx,dx
	_repeat 16
		mov	al,PortReg
		res	al,P_DAT+S_DAT+S_CLK
		set	al,P_CLK
		outp	CtrlPort,al
		res	al,P_CLK
		outp	CtrlPort,al
	 _if ne, dx,40h		; IBMPC
		inc	dx
		mov	ah,BIT7
	 _else			; PC98
		incw	dx
		mov	ah,BIT2
	 _endif
		in	al,dx
		test	al,ah
	 @if z, stc
		rcr	bx,1
	_loop
		mov	ax,bx
		and	ax,0111111111111110b
		shr	ax,1
		pop	dx,cx,bx
		ret
RCVR_DATA	endp


;--------------------------------------;
; Device power supply operation

VCC_ON		proc
		mov	Ctrl1,00000000b
		mov	Ctrl2,01110100b
		 call	SET_REG
		 call	DLY10
		ret
VCC_ON		endp


VCC_OFF		proc
		mov	Ctrl1,00000000b
		mov	Ctrl2,00111100b
		mov	PortReg,0000000b
		 call	SET_REG
		 call	DLY
		ret
VCC_OFF		endp


VPP_ON		proc
		set	PortReg,P_DAT
		res	PortReg,P_CLK
		 call	SET_REG
		mov	Ctrl2,01010100b
		 call	SET_REG
		ret
VPP_ON		endp


VPP_OFF		proc
		mov	Ctrl2,01110100b
		 call	SET_REG
		ret
VPP_OFF		endp


; --------------------------------------;
; Setting of PGAL of the shift register
;
; Ctrl1, to transfer the value of the Ctrl2

SET_REG		proc
		push	ax,bx,cx,dx
		mov	dx,CtrlPort
		mov	al,PortReg
		mov	bh,Ctrl2	; send U1/U2 data
		mov	bl,Ctrl1
	_repeat 16
		res	al,S_DAT+S_CLK
		shl	bx,1
	 @if c,<set	al,S_DAT>
		out	dx,al
		set	al,S_CLK
		out	dx,al
	_loop
		set	al,S_STB	; STB
		out	dx,al		;
		res	al,S_STB+S_CLK	;
		out	dx,al		;
		mov	PortReg,al
		pop	dx,cx,bx,ax
		ret
SET_REG		endp


;--------------------------------------;
; 10 ms waiting

DLY10		proc
		push	ax,dx
		set	PortReg,T_TRG+S_CLK	; trigger timer
		outb	CtrlPort,PortReg
		res	PortReg,T_TRG+S_DAT
		outb	CtrlPort,PortReg

	_if ne, dx,40h		; IBMPC
		inc	dx
		mov	ah,BIT7
	_else			; PC98
		mov	dx,42h
		mov	ah,BIT2
	_endif
	_while
		in	al,dx
	_rept z,<test	al,ah>
		pop	dx,ax
		ret
DLY10		endp


;--------------------------------------;
; Number s wait (I / O access use)

DLY		proc
		push	ax,dx
		inp	al,CtrlPort
		inp	al,CtrlPort
		inp	al,CtrlPort
		pop	dx,ax
		ret
DLY		endp


;--------------------------------------;

		$end	start

